/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program and library             */
/*         SCIP --- Solving Constraint Integer Programs                      */
/*                                                                           */
/*    Copyright (C) 2002-2009 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SCIP is distributed under the terms of the ZIB Academic License.         */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SCIP; see the file COPYING. If not email to scip@zib.de.      */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#pragma ident "@(#) $Id: Exp $"

/**@file   pricer_binpacking.c
 * @brief  binpacking variable pricer
 * @author Stefan Heinz
 */

/*---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8----+----9----+----0----+----1----+----2*/

#include <assert.h>
#include <string.h>

#include "pricer_binpacking.h"
#include "scip/cons_knapsack.h"
#include "scip/cons_setppc.h"

#define PRICER_NAME            "binpacking"
#define PRICER_DESC            "pricer for binpacking tours"
#define PRICER_PRIORITY        0
#define PRICER_DELAY           TRUE     /* only call pricer if all problem variables have non-negative reduced costs */




/*
 * Data structures
 */

/** variable pricer data */
struct SCIP_PricerData
{
   SCIP_CONS** conss;                        /**< set covering constraints for the items */
   SCIP_Longint* weights;                    /**< size of the items */
   int nweights;                             /**< number of weights (items) to be packed */
   SCIP_Real capacity;                       /**< capacity of the bins */
};


/*
 * Callback methods of variable pricer
 */


/** destructor of variable pricer to free user data (called when SCIP is exiting) */
static
SCIP_DECL_PRICERFREE(pricerFreeBinpacking)
{
   SCIP_PRICERDATA* pricerdata;
  
   assert(scip != NULL);
   assert(pricer != NULL);
   
   pricerdata = SCIPpricerGetData(pricer);
 
   if( pricerdata != NULL)
   {
      /* free memory */
      SCIPfreeMemoryArray(scip, &pricerdata->conss);
      SCIPfreeMemoryArray(scip, &pricerdata->weights);
      
      SCIPfreeMemoryNull(scip, &pricerdata);
   }
   
   return SCIP_OKAY;
}


/** initialization method of variable pricer (called after problem was transformed) */
static
SCIP_DECL_PRICERINIT(pricerInitBinpacking)
{  /*lint --e{715}*/
   SCIP_PRICERDATA* pricerdata;
   SCIP_VAR** vars;
   SCIP_CONS* cons;
   int nvars;
   int c;

   assert(scip != NULL);
   assert(pricer != NULL);
  
   pricerdata = SCIPpricerGetData(pricer);
   assert(pricerdata != NULL);
  
   /* get transformed constraints */
   for( c = 0; c < pricerdata->nweights; ++c )
   {
      cons = pricerdata->conss[c];

      /* release original constraint */
      SCIP_CALL( SCIPreleaseCons(scip, &pricerdata->conss[c]) );

      /* get transformed constraint */
      SCIP_CALL( SCIPgetTransformedCons(scip, cons, &pricerdata->conss[c]) );

      /* capture transformed constraint */
      SCIP_CALL( SCIPcaptureCons(scip, pricerdata->conss[c]) );
   }

   /* set the upper bound for each decision variable to lazy since these upper bounds are enforced through the set
    * packing constraints */
   vars = SCIPgetVars(scip);
   nvars = SCIPgetNVars(scip);
   for( c = 0; c < nvars; ++c )
   {
      SCIP_CALL( SCIPchgVarUbLazy(scip, vars[c], 1.0) );
   }

   return SCIP_OKAY;
}


/** deinitialization method of variable pricer (called before transformed problem is freed) */
#define pricerExitBinpacking NULL


/** solving process initialization method of variable pricer (called when branch and bound process is about to begin) */
#define pricerInitsolBinpacking NULL


/** solving process deinitialization method of variable pricer (called before branch and bound process data is freed) */
static
SCIP_DECL_PRICEREXITSOL(pricerExitsolBinpacking)
{
   SCIP_PRICERDATA* pricerdata;
   int c;
  
   assert(scip != NULL);
   assert(pricer != NULL);

   pricerdata = SCIPpricerGetData(pricer);
   assert(pricerdata != NULL);
  
   /* get release constraints */
   for( c = 0; c < pricerdata->nweights; ++c )
   {
      /* release constraint */
      SCIP_CALL( SCIPreleaseCons(scip, &(pricerdata->conss[c])) );
   }
   
   return SCIP_OKAY;
}


/** reduced cost pricing method of variable pricer for feasible LPs */
static
SCIP_DECL_PRICERREDCOST(pricerRedcostBinpacking)
{  /*lint --e{715}*/
   SCIP_PRICERDATA* pricerdata;
   SCIP_CONS** conss;
   SCIP_Longint* weights;
   SCIP_Real* profits;
   int* solitems;
   int* nonsolitems;
   int* items;
   SCIP_Real redcost;
   
   char varname[SCIP_MAXSTRLEN];
   
   int nweights;
   int nsolitems;
   int nnonsolitems;
   SCIP_Longint capacity;
   int i;
   
   assert(scip != NULL);
   assert(pricer != NULL);
   
   /* get the pricer data */
   pricerdata = SCIPpricerGetData(pricer);
   assert(pricerdata != NULL);
   
   capacity = pricerdata->capacity;
   conss = pricerdata->conss;
   nweights = pricerdata->nweights;

   /* allocated memory for profit array, solution array, non solution array, and items array */
   SCIP_CALL( SCIPallocBufferArray(scip, &items, nweights) );
   SCIP_CALL( SCIPallocBufferArray(scip, &profits, nweights) );
   SCIP_CALL( SCIPallocBufferArray(scip, &solitems, nweights) );
   SCIP_CALL( SCIPallocBufferArray(scip, &nonsolitems, nweights) );

   /* copy the weights array */
   SCIP_CALL( SCIPduplicateBufferArray(scip, &weights, pricerdata->weights, nweights) );
   
   nsolitems = 0;
   nnonsolitems = 0;

   /* fill profit array with dual solution value */
   for( i = 0; i < nweights; i++ ) 
   {
      profits[i] = SCIPgetDualsolSetppc(scip, conss[i]);
      items[i] = i;
   }
   
#ifdef SCIP_DEBUG
   /* debug output; this one turned on using "#define SCIP_DEBUG" */

   SCIPdebugMessage("   nitems: %d capacity: %"SCIP_LONGINT_FORMAT"\n", nweights, capacity);
   SCIPdebugMessage(" %4s %4s %7s\n", "#", "dual", "weights");
   for( i = 0; i < nweights; ++i ) 
   {
      SCIPdebugMessage("%4d %4.2f %7"SCIP_LONGINT_FORMAT"\n", items[i], profits[i], weights[i]);
   }
#endif
   
   /* solve the knapsack sack problem */
   SCIP_CALL( SCIPsolveKnapsackExactly(scip, nweights, weights, profits, capacity, items, 
         solitems, nonsolitems, &nsolitems, &nnonsolitems, &redcost) );

#ifdef SCIP_DEBUG
   /* debug output; this one turned on using "#define SCIP_DEBUG" */

   SCIPdebugMessage("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");
   SCIPdebugMessage("X Knapsack algorithm calculated reduced costs of %f X \n",redcost);
   SCIPdebugMessage("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");   

   {
      int item;
      SCIP_Longint usedcapacity;

      usedcapacity = 0;
      
      SCIPdebugMessage(" nitems: %d capacity: %"SCIP_LONGINT_FORMAT"\n", nweights, capacity);
      SCIPdebugMessage(" %4s %7s\n", "#", "weights");
      for( i = 0; i < nsolitems; ++i ) 
      {
         item  = solitems[i];
         usedcapacity += weights[item];
         SCIPdebugMessage(" %4d %7"SCIP_LONGINT_FORMAT"\n", item, pricerdata->weights[item]);
      }
      assert(usedcapacity <= capacity);
   }
#endif
   
   /* check if we found a solution with negative reduced cost */
   if( SCIPisFeasGT(scip, redcost, 1.0) ) 
   {
      SCIP_VAR* var;
      char strtmp[SCIP_MAXSTRLEN];
      
      /* create a suitable variable name; this might help for debugging */
      (void) SCIPsnprintf(varname, SCIP_MAXSTRLEN, "items"); 
      
      for( i = 0; i < nsolitems; ++i )
      {
         (void) SCIPsnprintf(strtmp, SCIP_MAXSTRLEN, "_%d", solitems[i]); 
         strcat(varname, strtmp); 
      }
      
      /* create a new binary variable */
      SCIP_CALL( SCIPcreateVar(scip, &var, varname, 0.0, 1.0, 1.0, SCIP_VARTYPE_BINARY, 
            TRUE, TRUE, NULL, NULL, NULL, NULL) );

      /* add the new variable to the pricer store */
      SCIP_CALL( SCIPaddPricedVar(scip, var, 1.0) );

      /* change the upper bound of the binary variable to lazy since the upper bound is already enforced 
       * due to the objective function the set covering constraint;
       * The reason for doing is that, is to avoid the bound of x <= 1 in the LP relaxation since this bound
       * constraint would produce a dual variable which might have a positive reduced cost 
       */
      SCIP_CALL( SCIPchgVarUbLazy(scip, var, 1.0) );

      /* add the new variable to the corresponding set covering constarints */
      for( i = 0; i < nsolitems; ++i )
      {
         SCIP_CALL( SCIPaddCoefSetppc(scip, conss[solitems[i]], var) );
      }
      
      /* release the new variable such that SCIP takes care it */
      SCIP_CALL( SCIPreleaseVar(scip, &var) );      
   }
   
   /* free buffer arrarys */
   SCIPfreeBufferArray(scip, &weights);
   SCIPfreeBufferArray(scip, &nonsolitems);
   SCIPfreeBufferArray(scip, &solitems);
   SCIPfreeBufferArray(scip, &profits);
   SCIPfreeBufferArray(scip, &items);

   (*result) = SCIP_SUCCESS;

   return SCIP_OKAY;
}


#if 0
/** farkas pricing method of variable pricer for infeasible LPs */
static
SCIP_DECL_PRICERFARKAS(pricerFarkasBinpacking)
{  /*lint --e{715}*/
   SCIPerrorMessage("method of binpacking variable pricer not implemented yet\n");
   SCIPABORT(); /*lint --e{527}*/

   return SCIP_OKAY;
}
#else
#define pricerFarkasBinpacking NULL
#endif




/*
 * variable pricer specific interface methods
 */

/** creates the binpacking variable pricer and includes it in SCIP */
SCIP_RETCODE SCIPincludePricerBinpacking(
   SCIP*                 scip                /**< SCIP data structure */
   )
{
   SCIP_PRICERDATA* pricerdata;

   /* create binpacking variable pricer data */
   SCIP_CALL( SCIPallocMemory(scip, &pricerdata) );

   pricerdata->conss = NULL;
   pricerdata->weights = NULL;
   pricerdata->nweights = 0;
   pricerdata->capacity = 0.0;
   
   /* include variable pricer */
   SCIP_CALL( SCIPincludePricer(scip, PRICER_NAME, PRICER_DESC, PRICER_PRIORITY, PRICER_DELAY,
         pricerFreeBinpacking, pricerInitBinpacking, pricerExitBinpacking, 
         pricerInitsolBinpacking, pricerExitsolBinpacking, pricerRedcostBinpacking, pricerFarkasBinpacking,
         pricerdata) );

   /* add binpacking variable pricer parameters */
   /* TODO: (optional) add variable pricer specific parameters with SCIPaddTypeParam() here */

   return SCIP_OKAY;
}


/** added problem specific data to pricer and activates pricer */
SCIP_RETCODE SCIPpricerBinpackingActivate(
   SCIP*                 scip,               /**< SCIP data structure */
   SCIP_CONS**           conss,              /**< set covering constraints for the items */
   SCIP_Longint*         weights,            /**< size of the items */
   int                   nweights,           /**< number of weights (items) to be packed */
   SCIP_Real             capacity            /**< capacity of the bins */
   )
{
   SCIP_PRICER* pricer;
   SCIP_PRICERDATA* pricerdata;
   int c;

   assert(scip != NULL);
   assert(conss != NULL);
   assert(weights != NULL);
   assert(nweights > 0);
   
   pricer = SCIPfindPricer(scip, PRICER_NAME);
   assert(pricer != NULL);
  
   pricerdata = SCIPpricerGetData(pricer);
   assert(pricerdata != NULL);
   
   /* copy arrays */
   SCIP_CALL( SCIPduplicateMemoryArray(scip, &pricerdata->conss, conss, nweights) );
   SCIP_CALL( SCIPduplicateMemoryArray(scip, &pricerdata->weights, weights, nweights) );

   pricerdata->nweights = nweights;
   pricerdata->capacity = capacity;

   SCIPdebugMessage("   nitems: %d capacity: %f  \n", nweights, capacity);
   SCIPdebugMessage("      # profits    weights   x  \n");   /* capture constraints */
   
   /* capture all constraints */
   for( c = 0; c < nweights; ++c )
   {
      SCIP_CALL( SCIPcaptureCons(scip, conss[c]) );
      SCIPdebugPrintf("%4d %3"SCIP_LONGINT_FORMAT"\n", c, weights[c]);
   }
   
   /* activate pricer */
   SCIP_CALL( SCIPactivatePricer(scip, pricer) );

   return SCIP_OKAY;
}
